@@ -21,7 +21,7 @@
 Frescobaldi Main Window.
 """
 
-from __future__ import division
+
 
 import itertools
 import os
@@ -34,7 +34,7 @@ from PyQt5.QtGui import (QKeySequence, QTextCursor, QTextDocument)
 from PyQt5.QtPrintSupport import (QAbstractPrintDialog, QPrintDialog, QPrinter)
 from PyQt5.QtWidgets import (QAction, QApplication, QFileDialog, QMainWindow,
                              QMenu, QMessageBox, QPlainTextEdit, QVBoxLayout,
-                             QWhatsThis, QWidget)
+                             QWhatsThis, QWidget, QInputDialog)
 
 import app
 import backup
@@ -57,9 +57,9 @@ import helpers
 import panelmanager
 import engrave
 import scorewiz
+import snippet
 import externalchanges
 import browseriface
-import vcs
 import file_import
 
 
@@ -124,6 +124,7 @@ class MainWindow(QMainWindow):
         layout.addWidget(self.tabBar)
         layout.addWidget(self.viewManager)
 
+        app.load_extensions(self)
         self.createActions()
         self.createMenus()
         self.createToolBars()
@@ -146,6 +147,8 @@ class MainWindow(QMainWindow):
             self.setCurrentDocument(other.currentDocument())
         self.updateWindowTitle()
         app.mainwindowCreated(self)
+        app.settingsChanged.connect(self.settingsChanged)
+        self.settingsChanged()
 
     def documents(self):
         """Returns the list of documents in the order of the TabBar."""
@@ -215,15 +218,18 @@ class MainWindow(QMainWindow):
                 curd.undoAvailable.disconnect(self.updateDocActions)
                 curd.redoAvailable.disconnect(self.updateDocActions)
                 curd.modificationChanged.disconnect(self.updateWindowTitle)
+                curd.urlChanged.disconnect(self.updateFileActions)
                 curd.urlChanged.disconnect(self.updateWindowTitle)
                 curd.loaded.disconnect(self.updateDocActions)
             doc.undoAvailable.connect(self.updateDocActions)
             doc.redoAvailable.connect(self.updateDocActions)
             doc.modificationChanged.connect(self.updateWindowTitle)
+            doc.urlChanged.connect(self.updateFileActions)
             doc.urlChanged.connect(self.updateWindowTitle)
             doc.loaded.connect(self.updateDocActions)
             self.updateDocActions()
             self.updateWindowTitle()
+            self.updateFileActions()
         self.updateSelection()
         self.updateActions()
         self.currentViewChanged.emit(view, curv)
@@ -246,6 +252,17 @@ class MainWindow(QMainWindow):
         action = self.actionCollection.view_wrap_lines
         action.setChecked(view.lineWrapMode() == QPlainTextEdit.WidgetWidth)
 
+    def updateFileActions(self):
+        doc = self.currentDocument()
+        ac = self.actionCollection.file_rename
+        if doc:
+            if doc.url().isEmpty():
+                ac.setEnabled(False)
+            elif doc.url().toLocalFile():
+                ac.setEnabled(True)
+            else:
+                ac.setEnabled(False)
+
     def updateDocActions(self):
         doc = self.currentDocument()
         ac = self.actionCollection
@@ -270,7 +287,8 @@ class MainWindow(QMainWindow):
 
         window_title = app.caption(" ".join(name))
 
-        if vcs.app_is_git_controlled():
+        if app.is_git_controlled():
+            import vcs
             window_title += " " + vcs.app_active_branch_window_title()
 
         self.setWindowTitle(window_title)
@@ -282,7 +300,7 @@ class MainWindow(QMainWindow):
             impurls = []
             for url in ev.mimeData().urls():
                 imp = file_import.FileImport.instance(self)
-                if imp.isImportable(url.toLocalFile()):
+                if imp.is_importable(url.toLocalFile()):
                     impurls.append(QDir.toNativeSeparators(url.toLocalFile()))
                 else:
                     lyurls.append(url)
@@ -290,7 +308,8 @@ class MainWindow(QMainWindow):
             if docs:
                 self.setCurrentDocument(docs[-1])
             for i in impurls:
-                imp.openDialog(i)
+                imp.configure_import(i)
+                imp.run_import()
 
     def dragEnterEvent(self, ev):
         if not ev.source() and ev.mimeData().hasUrls():
@@ -337,7 +356,7 @@ class MainWindow(QMainWindow):
             else:
                 allow_close = res == QMessageBox.Discard
         return allow_close and engrave.engraver(self).queryCloseDocument(doc)
-    
+
     def createPopupMenu(self):
         """ Adds an entry to the popup menu to show/hide the tab bar. """
         menu = QMainWindow.createPopupMenu(self)
@@ -476,7 +495,7 @@ class MainWindow(QMainWindow):
 
     def newDocument(self):
         """ Creates a new, empty document. """
-        d = document.Document()
+        d = document.EditorDocument()
         self.setCurrentDocument(d)
         s = QSettings()
         ndoc = s.value("new_document", "empty", str)
@@ -510,6 +529,32 @@ class MainWindow(QMainWindow):
         if docs:
             self.setCurrentDocument(docs[-1])
 
+    def renameDocument(self, doc):
+        url = doc.url()
+        filename = url.toLocalFile()
+        filetypes = app.filetypes(os.path.splitext(filename)[1])
+        caption = app.caption(_("dialog title", "Rename/Move File"))
+        new_filename = QFileDialog.getSaveFileName(self, caption, filename, filetypes)[0]
+        if not new_filename or filename == new_filename:
+            return False # cancelled
+        new_url = QUrl.fromLocalFile(new_filename)
+        doc.setUrl(new_url)
+        if self.saveDocument(doc):
+            try:
+                os.remove(filename)
+            except IOError as e:
+                msg = _("{message}\n\n{strerror} ({errno})").format(
+                    message = _("Could not delete: {url}").format(url=filename),
+                    strerror = e.strerror,
+                    errno = e.errno)
+                QMessageBox.critical(self, app.caption(_("Error")), msg)
+                return False
+            else:
+                recentfiles.remove(url)
+            return True
+        else:
+            return False
+
     def saveDocument(self, doc, save_as=False):
         """ Saves the document, asking for a name if necessary.
 
@@ -529,7 +574,7 @@ class MainWindow(QMainWindow):
                         break
                 else:
                     directory = app.basedir() # default directory to save to
-                
+
                 import documentinfo
                 import ly.lex
                 filename = os.path.join(directory, documentinfo.defaultfilename(doc))
@@ -586,6 +631,9 @@ class MainWindow(QMainWindow):
                 self.cleanStart()
         return close
 
+    def renameCurrentDocument(self):
+        return self.renameDocument(self.currentDocument())
+
     def saveCurrentDocument(self):
         return self.saveDocument(self.currentDocument())
 
@@ -678,14 +726,10 @@ class MainWindow(QMainWindow):
         self.setCurrentDocument(cur, findOpenView=True)
         return True
 
-    def closeOtherDocuments(self):
-        """ Closes all documents that are not the current document.
-
-        Returns True if all documents were closed.
-
-        """
+    def _closeDocuments(self, current=True):
         cur = self.currentDocument()
-        docs = self.historyManager.documents()[1:]
+        start = 0 if current else 1
+        docs = self.historyManager.documents()[start:]
         for doc in docs:
             if not self.queryCloseDocument(doc):
                 self.setCurrentDocument(cur, findOpenView=True)
@@ -694,7 +738,22 @@ class MainWindow(QMainWindow):
             doc.close()
         return True
 
+    def closeOtherDocuments(self):
+        """ Closes all documents that are not the current document.
+
+        Returns True if all documents were closed.
+
+        """
+        return self._closeDocuments(current=False)
+
     def closeAllDocuments(self):
+        """Closes all documents and keep one new, empty document
+        within the current session."""
+        if self._closeDocuments(current=True):
+            self.allDocumentsClosed.emit()
+            self.cleanStart()
+
+    def closeAllDocumentsAndSession(self):
         """Closes all documents and keep one new, empty document."""
         sessions.manager.get(self).saveCurrentSessionIfDesired()
         if self.queryClose():
@@ -900,6 +959,37 @@ class MainWindow(QMainWindow):
         sb = self.currentView().verticalScrollBar()
         sb.setValue(sb.value() + 1)
 
+    def gotoLine(self):
+        """Ask for line number and go there"""
+        line_count = self.currentDocument().blockCount()
+        view = self.currentView()
+        cur = view.textCursor()
+        current_block = cur.block()
+        current_line = current_block.firstLineNumber()
+        char_pos = cur.position() - current_block.position()
+        loc_pos = view.cursorRect(cur).bottomLeft()
+        pos = view.viewport().mapToGlobal(loc_pos)
+
+        dlg = QInputDialog(self)
+        dlg.setInputMode(QInputDialog.IntInput)
+        dlg.setIntMinimum(1)
+        dlg.setIntMaximum(line_count)
+        dlg.setIntValue(current_line)
+        dlg.setLabelText(_("Goto Line Number (1-{num}):").format(num=line_count))
+        dlg.setWindowFlags(Qt.Popup)
+        dlg.move(pos)
+        dlg_result = dlg.exec()
+        if dlg_result:
+            line = dlg.intValue()
+            cur = QTextCursor(self.currentDocument().findBlockByNumber(line - 1))
+            new_block = cur.block()
+            if new_block.length() > char_pos:
+                cur.setPosition(cur.position() + char_pos)
+            else:
+                cur.setPosition(cur.position() + new_block.length() - 1)
+            view.setTextCursor(cur)
+            view.centerCursor()
+
     def selectFullLinesUp(self):
         """Select lines upwards, selecting full lines."""
         self.selectFullLines(QTextCursor.Up)
@@ -957,6 +1047,7 @@ class MainWindow(QMainWindow):
         ac.file_save.triggered.connect(self.saveCurrentDocument)
         ac.file_save_as.triggered.connect(self.saveCurrentDocumentAs)
         ac.file_save_copy_as.triggered.connect(self.saveCopyAs)
+        ac.file_rename.triggered.connect(self.renameCurrentDocument)
         ac.file_save_all.triggered.connect(self.saveAllDocuments)
         ac.file_reload.triggered.connect(self.reloadCurrentDocument)
         ac.file_reload_all.triggered.connect(self.reloadAllDocuments)
@@ -965,6 +1056,9 @@ class MainWindow(QMainWindow):
         ac.file_close.triggered.connect(self.closeCurrentDocument)
         ac.file_close_other.triggered.connect(self.closeOtherDocuments)
         ac.file_close_all.triggered.connect(self.closeAllDocuments)
+        ac.file_close_all_and_session.triggered.connect(
+            self.closeAllDocumentsAndSession
+        )
         ac.export_colored_html.triggered.connect(self.exportColoredHtml)
         ac.edit_undo.triggered.connect(self.undo)
         ac.edit_redo.triggered.connect(self.redo)
@@ -985,6 +1079,7 @@ class MainWindow(QMainWindow):
         ac.view_wrap_lines.triggered.connect(self.toggleWrapLines)
         ac.view_scroll_up.triggered.connect(self.scrollUp)
         ac.view_scroll_down.triggered.connect(self.scrollDown)
+        ac.view_goto_line.triggered.connect(self.gotoLine)
         ac.window_new.triggered.connect(self.newWindow)
         ac.window_fullscreen.toggled.connect(self.toggleFullScreen)
         ac.help_manual.triggered.connect(self.showManual)
@@ -1015,15 +1110,28 @@ class MainWindow(QMainWindow):
         self.addAction(ac.edit_select_full_lines_up)
         self.addAction(ac.edit_select_full_lines_down)
 
+    def settingsChanged(self):
+        ac = self.actionCollection
+        t = self.toolbar_main
+        if QSettings().value("verbose_toolbuttons", False, bool):
+            import snippet.menu
+            new = snippet.menu.TemplateMenu(self)
+            save = menu.menu_file_save(self)
+            close = menu.menu_file_close(self)
+        else:
+            new = save = close = None
+        t.widgetForAction(ac.file_new).setMenu(new)
+        t.widgetForAction(ac.file_save).setMenu(save)
+        t.widgetForAction(ac.file_close).setMenu(close)
+
     def createToolBars(self):
         ac = self.actionCollection
         self.toolbar_main = t = self.addToolBar('')
         t.setObjectName('toolbar_main')
         t.addAction(ac.file_new)
         t.addAction(ac.file_open)
+        t.widgetForAction(ac.file_open).setMenu(self.menu_recent_files)
         t.addAction(ac.file_save)
-        t.widgetForAction(ac.file_save).addAction(ac.file_save_as)
-        t.widgetForAction(ac.file_save).addAction(ac.file_save_all)
         t.addAction(ac.file_close)
         t.addSeparator()
         t.addAction(browseriface.get(self).actionCollection.go_back)
@@ -1047,6 +1155,7 @@ class MainWindow(QMainWindow):
         t.addAction(ma.music_zoom_in)
         t.addAction(ma.music_zoom_combo)
         t.addAction(ma.music_zoom_out)
+        t.addAction(ma.music_magnifier)
         t.addSeparator()
         t.addAction(ma.music_prev_page)
         t.addAction(ma.music_pager)
@@ -1069,6 +1178,7 @@ class ActionCollection(actioncollection.ActionCollection):
         self.file_save = QAction(parent)
         self.file_save_as = QAction(parent)
         self.file_save_copy_as = QAction(parent)
+        self.file_rename = QAction(parent)
         self.file_save_all = QAction(parent)
         self.file_reload = QAction(parent)
         self.file_reload_all = QAction(parent)
@@ -1077,6 +1187,7 @@ class ActionCollection(actioncollection.ActionCollection):
         self.file_close = QAction(parent)
         self.file_close_other = QAction(parent)
         self.file_close_all = QAction(parent)
+        self.file_close_all_and_session = QAction(parent)
         self.file_quit = QAction(parent)
         self.file_restart = QAction(parent)
 
@@ -1104,6 +1215,7 @@ class ActionCollection(actioncollection.ActionCollection):
         self.view_wrap_lines = QAction(parent, checkable=True)
         self.view_scroll_up = QAction(parent)
         self.view_scroll_down = QAction(parent)
+        self.view_goto_line = QAction(parent)
 
         self.window_new = QAction(parent)
         self.window_fullscreen = QAction(parent)
@@ -1123,6 +1235,7 @@ class ActionCollection(actioncollection.ActionCollection):
         self.file_save.setIcon(icons.get('document-save'))
         self.file_save_as.setIcon(icons.get('document-save-as'))
         self.file_save_copy_as.setIcon(icons.get('document-save-as'))
+        self.file_rename.setIcon(icons.get('document-save-as'))
         self.file_save_all.setIcon(icons.get('document-save-all'))
         self.file_reload.setIcon(icons.get('reload'))
         self.file_reload_all.setIcon(icons.get('reload-all'))
@@ -1178,10 +1291,15 @@ class ActionCollection(actioncollection.ActionCollection):
         self.edit_replace.setShortcuts(QKeySequence.Replace)
         self.edit_preferences.setShortcuts(QKeySequence.Preferences)
 
-        self.view_next_document.setShortcuts(QKeySequence.Forward)
-        self.view_previous_document.setShortcuts(QKeySequence.Back)
+        if sys.platform.startswith('darwin'):
+            self.view_next_document.setShortcut(Qt.META + Qt.Key_Tab)
+            self.view_previous_document.setShortcut(Qt.META + Qt.SHIFT + Qt.Key_Tab)
+        else:
+            self.view_next_document.setShortcuts(QKeySequence.Forward)
+            self.view_previous_document.setShortcuts(QKeySequence.Back)
         self.view_scroll_up.setShortcut(Qt.CTRL + Qt.Key_Up)
         self.view_scroll_down.setShortcut(Qt.CTRL + Qt.Key_Down)
+        self.view_goto_line.setShortcut(Qt.CTRL + Qt.ALT + Qt.Key_G)
 
         self.window_fullscreen.setShortcuts([QKeySequence(Qt.CTRL + Qt.SHIFT + Qt.Key_F), QKeySequence(Qt.Key_F11)])
 
@@ -1200,15 +1318,20 @@ class ActionCollection(actioncollection.ActionCollection):
                 self.help_about.setMenuRole(QAction.NoRole)
 
     def translateUI(self):
-        self.file_new.setText(_("action: new document", "&New"))
+        self.file_new.setText(_("action: new document", "&New Document"))
+        self.file_new.setIconText(_("action: new document", "New"))
         self.file_open.setText(_("&Open..."))
         self.file_open_recent.setText(_("Open &Recent"))
         self.file_insert_file.setText(_("Insert from &File..."))
+        self.file_insert_file.setToolTip(
+            _("Insert the contents of a file at the current cursor position."))
         self.file_open_current_directory.setText(_("Open Current Directory"))
         self.file_open_command_prompt.setText(_("Open Command Prompt"))
-        self.file_save.setText(_("&Save"))
+        self.file_save.setText(_("&Save Document"))
+        self.file_save.setIconText(_("Save"))
         self.file_save_as.setText(_("Save &As..."))
         self.file_save_copy_as.setText(_("Save Copy or Selection As..."))
+        self.file_rename.setText(_("&Rename/Move File..."))
         self.file_save_all.setText(_("Save All"))
         self.file_reload.setText(_("Re&load"))
         self.file_reload_all.setText(_("Reload All"))
@@ -1217,10 +1340,19 @@ class ActionCollection(actioncollection.ActionCollection):
             "Opens a window to check whether open documents were changed or "
             "deleted by other programs."))
         self.file_print_source.setText(_("Print Source..."))
-        self.file_close.setText(_("&Close"))
+        self.file_close.setText(_("&Close Document"))
+        self.file_close.setIconText(_("Close"))
         self.file_close_other.setText(_("Close Other Documents"))
-        self.file_close_all.setText(_("Close All Documents and Session"))
-        self.file_close_all.setToolTip(_("Closes all documents and leaves the current session."))
+        self.file_close_all.setText(_("Close All Documents"))
+        self.file_close_all.setToolTip(
+            _("Closes all documents but preserves the current session.")
+        )
+        self.file_close_all_and_session.setText(
+            _("Close All Documents and Session")
+        )
+        self.file_close_all_and_session.setToolTip(
+            _("Closes all documents and leaves the current session.")
+        )
         self.file_quit.setText(_("&Quit"))
         self.file_restart.setText(_("Restart {appname}").format(appname=appinfo.appname))
 
@@ -1248,6 +1380,7 @@ class ActionCollection(actioncollection.ActionCollection):
         self.view_wrap_lines.setText(_("Wrap &Lines"))
         self.view_scroll_up.setText(_("Scroll Up"))
         self.view_scroll_down.setText(_("Scroll Down"))
+        self.view_goto_line.setText(_("&Goto Line..."))
 
         self.window_new.setText(_("New &Window"))
         self.window_fullscreen.setText(_("&Fullscreen"))
